"""
Created on Tue Dec 12 2017
Modified on Feb 06 2018

@author: johana
Genetic Fuzzy System - Agent Library
"""
# Libraries:
import random
import numpy as np
import skfuzzy as fuzz
import time

# Global variables: 

# Fuzzy logic
NBS = 5
NBV = 3
NBO = 4
INP = 4 #4
CIN = 6 # 8 Cones
OUT = 1
COU = 14 # 16 Cones
x_axis = np.arange(0, 2**NBS, 1)

STACA = 10
STAPN = 7
MEAN = 2
SKE = 4
KUR = 5
ENE = 6
HOMO = 8
UNCA1 = 1
UNCA2 = 4
UNCA3 = 7
UNCA4 = 10
NBPA1 = STAPN + 0
NBPA3 = STAPN + 6

# Camaras
INDSV = np.array([MEAN, UNCA2+STACA, UNCA3+STACA, UNCA4+STACA]) 
INDIR = np.array([HOMO, KUR, UNCA2+STACA, UNCA3+STACA])
INDUV = np.array([SKE, KUR, ENE, UNCA3+STACA])

# Ranges per sensor
MINSV = np.array([63,0,0,0])
MAXSV = np.array([125,1,1,1])
MINIR = np.array([0.1,-1.5,0,0])
MAXIR = np.array([0.6,0,1,1])
MINUV = np.array([-0.65,-1.5,0.01,0])
MAXUV = np.array([0.75,0,0.33,1])

MINACS = -30
MAXACS = 30 
# Otros
INDTM = np.array([UNCA1+STAPN, UNCA2+STAPN, UNCA3+STAPN, NBPA1])
INDGP = np.array([UNCA1+STAPN, UNCA2+STAPN, UNCA3+STAPN, UNCA4+STAPN])
MINTER = np.array([0, 0, 0, -30])
MAXTER = np.array([1, 1, 1, 30])
MINGPR = np.array([0, 0, 0, 0])
MAXGPR = np.array([1, 1, 1, 1])

MINAPS = -90
MAXAPS = 90

# FUZZY FUNCTIONS:

def fuzzyset(crisp,case = 0):
    az = 2**NBS
    nsetsi = CIN + 2
    nsetso = COU + 2
    delta = az/((COU+1)*2)
    if case == 0:
        result= np.zeros([INP,nsetsi,az])
        for j in range(INP):
            center = 0
            for k in range(nsetsi):
                if k == 0:
                    result[j,k,:] = fuzz.trimf(x_axis, [0, center, crisp[j,k]])
                elif 0 < k < nsetsi-2:
                    result[j,k,:] = fuzz.trimf(x_axis, [center, crisp[j,k-1], crisp[j,k]])
                    center=crisp[j,k-1]
                elif k == nsetsi-2:
                    result[j,k,:] = fuzz.trimf(x_axis, [center, crisp[j,k-1], az-1])
                    center=crisp[j,k-1]
                else:
                    result[j,k,:] =fuzz.trimf(x_axis, [center, az-1, az-1])
    else:
        result= np.zeros([OUT,nsetso,az])
        for j in range(OUT):
            center = 0
            for k in range(nsetso):
                if k == 0:
                    result[j,k,:] = fuzz.trimf(x_axis, [0, center, delta])
                elif 0 < k < nsetso-1:
                    center = crisp[j,k-1]
                    result[j,k,:] = fuzz.trimf(x_axis, [center-delta, center, center+delta])
                else:
                    center = az-1
                    result[j,k,:] =fuzz.trimf(x_axis, [center-delta, center, center])
    return result

def aplifuzzyset(dat,mmf): 
    return fuzz.interp_membership(x_axis, mmf, dat)

def evalfuzzysets(dat,setmf,nset):
    result = np.zeros(nset)    
    for i in range(nset):
        result[i] = aplifuzzyset(dat,setmf[i,:])
    return result

def appfuzzysample(sample,sett,nset):
    result = np.zeros([nset,INP]) 
    for i in range (INP):
        result[:,i] = evalfuzzysets(sample[i],sett[i,:,:],nset)
    return result

def appfuzzyrules(vlin,rules):
    nrul,var=rules.shape
    resulti = np.zeros(nrul)
    for i in range(nrul):
        resulti[i] = infuzzrule(vlin,rules[i,:])
    return resulti
        
def infuzzrule(vlin,rule):
    rulet = 0
    for j in range(INP):
        ind = int(rule[j]) 
        rulet = np.fmax(rulet, vlin[ind,j])
    return rulet

def outfuzzrule(rules,rin,sout):
    pout = 2**NBO
    a,b = sout.shape
    retout = np.zeros([pout,b])
    intout = np.zeros(pout)
    for j in range(pout):
        inds = rules[:,INP] == j # Restricted to one input       
        tmo = np.full_like(inds,1)
        tmot = tmo[inds]
        l = len(tmot)
        if l > 1:
            tempi = rin[inds]
            nrto = len(tempi)
            if nrto > 1:
                tempo = 0
                for k in range (nrto):
                    tempo = np.fmax(tempo,tempi[k])
                intout[j] = tempo
            else:
                intout[j] = tempi
    for k in range(pout):
        retout[k,:] = np.fmin(intout[k],sout[k,:])
    return retout

def fuzzaggregate(seto):
    opt,lon=seto.shape
    ruleo = np.zeros(lon)
    for j in range(opt):
        ruleo = np.fmax(ruleo,seto[j,:])
    return ruleo


# TOOLS FUNCTIONS:

# Dec to bin
def dec_to_bin(x,nb):
    ou=np.zeros(nb)    
    for i in range(nb):    
        r=x%2
        ou[nb-(i+1)]=r
        x=int(x/2)
    return ou
    
# Bin to dec
def bin_to_dec(x,nb):
    ou = 0    
    for i in range(nb):
        ind = nb-(i+1)
        ou = ou + x[ind]*2**i
    return ou

# Bin to dec array
def abin_to_adec(x,nb):
    ou=([])
    num = int (len(x)/nb)
    for i in range(num):
        ori = nb*i
        ou = np.append (ou,bin_to_dec(x[ori:ori+nb],nb))
    return ou


# Functions to apply fuzzy systems in agents
def SystemFuzzySets(indv,erul): 
    # Some useful sizes
    tam = len(indv)
    ten = len(erul)
    tsei = NBS*INP*CIN
    tseo = NBS*OUT*COU
    tcru = INP + OUT
    tstt = tsei+tseo    
    trul = tam-tstt
    
    # Dividing the information among sets and rules
    dsei = indv[0:tsei]  
    dseo = indv[tsei:tstt]
    drul = indv[tstt:tam]
    a = len(drul)
    col = INP*NBV + OUT*NBO
    row = int(a/col)    
    drul = np.resize(drul,[row,col])    
    
    # Transforming bin to dec
    sind = abin_to_adec(dsei,NBS)
    sopd = abin_to_adec(dseo,NBS)
    srud = ([])
    for k in range(row):
        auxi = abin_to_adec(drul[k,0:INP*NBV],NBV)
        auxo = abin_to_adec(drul[k,INP*NBV:col],NBO)
        srud = np.append(srud,auxi)
        srud = np.append(srud,auxo)
    
    # Ordered in an array
    sind = np.resize(sind,[INP,CIN])
    sopd = np.resize(sopd,[OUT,COU])
    srud = np.resize(srud,[ten,tcru])
    
    rr = 0    
    frul = ([])
    for i in range (max(INP,OUT,ten)):
        if i < INP:
            sind[i,:] = np.sort(sind[i,:])
        if i < OUT:
            sopd[i,:] = np.sort(sopd[i,:])
        if i < ten:
            if erul[i] == 1:
                frul = np.append(frul,srud[i,:])
                rr = rr+1
    frul = np.resize(frul,[rr,tcru])
    
    # Create fuzzy membership functions
    fsin = fuzzyset(sind,case = 0)
    fsou = fuzzyset(sopd,case = 1)
    return fsin,fsou,frul   
    
def FuzzyControlResult(fsin,fsou,frul,dati,outtype,tsensor):
    # Fuzzyfication
    lini = appfuzzysample(dati,fsin,CIN + 2)
    # Applying Fuzzy rules
    ouli = appfuzzyrules(lini,frul)
    lino = outfuzzrule(frul,ouli,fsou[0,:,:])
    outfuz = fuzzaggregate(lino)
    # Defuzzyfication
    fout = defuzzoutin(outfuz,outtype,tsensor)
    return fout

# return sample
def inputdata(tsensor,tem):
    result = np.zeros(INP)
    
    if (tsensor == 'A1'):
        ind = INDTM
        minv = MINTER
        maxv = MAXTER
    elif (tsensor == 'A2'):
        ind = INDGP
        minv = MINGPR
        maxv = MAXGPR    
    elif (tsensor == 'A3'):
        ind = INDSV
        minv = MINSV
        maxv = MAXSV 
    elif (tsensor == 'A4'):
        ind = INDIR
        minv = MINIR
        maxv = MAXIR        
    else:
        ind = INDUV
        minv = MINUV
        maxv = MAXUV 
    for k in range(INP):
        a = ind[k]
        result[k] = (2**NBS-1)*(tem[a] + abs(minv[k]))/(maxv[k]+abs(minv[k])) 
    return result

def defuzzoutin(outfuz,outtype,tsensor):
    outnor = fuzz.defuzz(x_axis, outfuz, 'centroid')
    if outtype == 'UNC':
        outf = outnor/(2**NBS-1)
    else:
        outf = 2*outnor/(2**NBS-1)-1
    return outf